/* graphics.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



/* All graphics drawing is done through subroutines in this module.
   This will compile for XFDI or GL.
*/




#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "graphics.h"
#include "matrix.h"
#include "x.h"



int Width, Height;        /* current window size */
int ScrWidth, ScrHeight;  /* screen size */
float AspectRatio;        /* window width/height aspect ratio */
int FontHeight;
int FontDescent;
int HQR_available;        /* High-Quality rendering available? */


static Window wmain;
static Matrix Ctm, CtmInv;
static Matrix IdMat = { 1.0, 0.0, 0.0, 0.0,
			0.0, 1.0, 0.0, 0.0,
                        0.0, 0.0, 1.0, 0.0,
			0.0, 0.0, 0.0, 1.0 };
static int pretty_flag = 0;


/******** STELLAR includes and variables *********/
#ifdef stellar
#  define DEFAULT_FONT "vr-25"
   static XFontStruct *visfont;
   static Display *dpy;
   static XFDIGC gc0;
   static GC gc1;
   static GC xgc;       /* extra GC for transparent surfaces */
   static XdbBufferAttributes  att;
   static unsigned long mask;
   static Drawable buff0, aabuff0, sorta, sorti, sortz;
   static float LightColor[3] = { 0.7, 0.7, 0.7 };
   static float AmbientColor[3] = { 1.0, 1.0, 1.0 };
   static float CurTrans;
   static float linewidth;
   /* Pretty rendering stuff: */
#  define AA_INC  (1.1/2)
   static float  xoffsets[AA_PASSES] =
     { 0, 0, 0, AA_INC, -AA_INC, AA_INC, -AA_INC, -AA_INC, AA_INC };
   static float  yoffsets[AA_PASSES] =
     { 0, -AA_INC, AA_INC, 0, 0, AA_INC, -AA_INC, AA_INC, -AA_INC };
   static float  blendcoeffs[AA_PASSES] =
     { 0.99999, 0.722/1.472, 0.722/2.194, 0.722/2.916, 0.722/3.638,
       0.639/4.277, 0.639/4.916, 0.639/5.555, 0.639/6.194 };
#endif


/******** SGI includes and variables *********/
#ifdef sgi
#  define sgi_or_ibm
#  include <fmclient.h>
#  define DEFAULT_FONT "Times-Roman"
#  define FLIP( A )  ( (A&0xff00ff00) | ((A>>16) & 0xff) | ((A&0xff) <<16) )
   static fmfonthandle visfont;
   static long zfar;
   static unsigned long CurColor = 0xffffffff;
   static unsigned long CurTrans = 0xff000000;
   static float front;
   static float xminplane[4], xmaxplane[4], yminplane[4], ymaxplane[4];

   /* setup the default material surface description: */
   static float mat1[] = {
      AMBIENT, 0.3, 0.3, 0.3,   /* surface ambient color */
      DIFFUSE, 0.7, 0.7, 0.7,   /* surface diffuse color */
      LMNULL
   };

   /* setup the light source descriptions: */
   static float light1[] = {
      LCOLOR, 0.8, 0.8, 0.8,          /* light color */
      POSITION, 0.0, 0.0, 10.0, 0.0,  /* light location */
      LMNULL
   };

   static float light2[] = {
      LCOLOR, 0.8, 0.8, 0.8,          /* light color */
      POSITION, 0.0, 0.0, -10.0, 0.0,  /* light location */
      LMNULL
   };

   /* setup the lighting model description: */
   static float lmodel[] = {
      AMBIENT, 0.2, 0.2, 0.2,         /* ambient color */
      LOCALVIEWER, 0,                 /* local viewpoint? */
      TWOSIDE, 0,
      LMNULL
   };

   /* Accumulation buffer antialiasing */
#  define AA_INC  1.0
   static float  xoffsets[AA_PASSES] =
     { -AA_INC, 0.0, AA_INC,  -AA_INC, 0.0, AA_INC,  -AA_INC, 0.0, AA_INC };
   static float  yoffsets[AA_PASSES] =
     { -AA_INC, -AA_INC, -AA_INC,  0.0, 0.0, 0.0,  AA_INC, AA_INC, AA_INC };

   static short screen1[16] = { /* 75% */
      0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd,
      0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd
     };
   static short screen2[16] = { /* 50% */
      0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555,
      0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555
     };
   static short screen3[16] = { /* 25% */
      0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222,
      0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222
     };
/*
   static char screen1[16][17] = { " XXX XXX XXX XXX",
				   "XX XXX XXX XXX X",
				   " XXX XXX XXX XXX",
				   "XX XXX XXX XXX X" };
   static char screen2[16][17] = { " X X X X X X X X",
                                   "X X X X X X X X ",
                                   " X X X X X X X X",
                                   "X X X X X X X X " };
   static char screen3[16][17] = { "X   X   X   X   ",
                                   "  X   X   X   X ",
				   "X   X   X   X   ",
                                   "  X   X   X   X " };
*/
   static int screendoor;

   static unsigned short  hourglass[16] = { 0x1ff0, 0x1ff0, 0x0820, 0x0820,
					    0x0820, 0x0c60, 0x06c0, 0x0100,
					    0x0100, 0x06c0, 0x0c60, 0x0820,
					    0x0820, 0x0820, 0x1ff0, 0x1ff0 };
#endif


/******** IBM includes and variables *********/
#ifdef ibm
#  define sgi_or_ibm
#  define FLIP( A )  ( (A&0xff00ff00) | ((A>>16) & 0xff) | ((A&0xff) <<16) )
#  define DOT( a, b )  ( a[0]*b[0] + a[1]*b[1] + a[2]*b[2] )
   static long zfar;
   static unsigned long CurColor = 0xffffffff;
   static unsigned long CurTrans = 0xff000000;
   static float front;
   static float xminplane[4], xmaxplane[4], yminplane[4], ymaxplane[4];

   /* setup the default material surface description: */
   static float mat1[] = {
      AMBIENT, 0.5, 0.5, 0.5,   /* surface ambient color */
      DIFFUSE, 0.7, 0.7, 0.7,   /* surface diffuse color */
      LMNULL
   };

   /* setup the light source descriptions: */
   static float light1[] = {
      LCOLOR, 0.6, 0.6, 0.6,          /* light color */
      POSITION, 0.0, 0.0, 10.0, 0.0,  /* light location */
      LMNULL
   };

   static float light2[] = {
      LCOLOR, 0.6, 0.6, 0.6,          /* light color */
      POSITION, 0.0, 0.0, -10.0, 0.0,  /* light location */
      LMNULL
   };

   /* setup the lighting model description: */
   static float lmodel[] = {
      AMBIENT, 0.4, 0.4, 0.4,         /* ambient color */
      LOCALVIEWER, 0,                 /* local viewpoint? */
      LMNULL
   };

   static short screen1[16] = { /* 75% */
      0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd,
      0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd, 0x7777, 0xdddd
     };
   static short screen2[16] = { /* 50% */
      0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555,
      0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555, 0xaaaa, 0x5555
     };
   static short screen3[16] = { /* 25% */
      0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222,
      0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222, 0x8888, 0x2222
     };
   static int screendoor;

#endif


/******** FOO includes and variables *********/
#ifdef foo
   /* include graphics library include file */
   /* declare local variables here. */
#endif




/*** init_graphics ****************************************************
   Initialize the graphics module.
   Input:  alphaflag - use alpha transparency flag
**********************************************************************/
void init_graphics( alphaflag )
int alphaflag;
{
#ifdef stellar
   ScrWidth = 1280;
   ScrHeight = 1024;
   HQR_available = 1;
#endif
#ifdef sgi
   ScrWidth = getgdesc( GD_XPMAX );
   ScrHeight = getgdesc( GD_YPMAX );
   if (getgdesc(GD_BITS_ACBUF)>0)
      HQR_available = 1;
   else
      HQR_available = 0;
   if (getgdesc(GD_BLEND) && alphaflag )
     screendoor = 0;
   else
     screendoor = 1;
#endif
#ifdef ibm
/* This bombs on some systems:
   ScrWidth = getgdesc( GD_XPMAX );
   ScrHeight = getgdesc( GD_YPMAX );
*/
   ScrWidth = 1024;
   ScrHeight = 768;
   HQR_available = 0;
   screendoor = 1;
#endif
#ifdef foo
   /* initialize any graphics stuff needed before opening the 3-D window */
   ScrWidth = 1024;  /* edit this */
   ScrHeight = 768;  /* edit this */
   HQR_available = 0;  /* don't need High-Quality rendering */
#endif
}




#ifdef sgi_or_ibm
/* find a window by name and return its window number */
static Window find_window( dpy, scr, start, name )
Display *dpy;
int scr;
Window start;
char *name;
{
   Status stat;
   int n, num;
   Window w, root, parent, (*children)[];
   char *title;

   if (XFetchName( dpy, start, &title )==1) {
      if (strcmp( name, title )==0) { 
         /* found it */
         XFree( title );
         return start;
      }
      XFree( title );
   }
 
   /* get list of child windows */
   stat = XQueryTree( dpy, start, &root, &parent, &children, &num );

   if (stat==1) {
      /* search each child and its offspring for the named window */
      for (n=0;n<num;n++) {
         w = find_window( dpy, scr, (*children)[n], name );
         if (w) {
	    XFree( children );
            return w;
         }
      }
      XFree( children );
   }

   /* not found */
   return 0;
}
#endif



/*** make_window ******************************************************
   Make a 3-D graphics window.
   Input:  display - which X display to open window on
           title - window title string
           xpos,ypos - position in pixel from upper-left corner
           width, height - width and height of window in pixels
           fontname - name of GL or X font to use
           fontsize - height of GL font to use
   Return:  X window pointer or NULL.
**********************************************************************/
Window make_window( display, title, xpos, ypos, width, height,
		    fontname, fontsize )
Display *display;
char *title;
int xpos, ypos;
int width, height;
char *fontname;
int fontsize;
{
#ifdef stellar
   int  scr;
   Screen  *screen;
   Visual  *visual;
   XColor  background, foreground;
   XSetWindowAttributes  xswa;
   XGCValues  gcv;
   XSizeHints sizehints;
   XEvent pe;
   int  depth;
   int  swa_flags;
   int  borderwidth;
   unsigned long  fg = 0xffffff;
   unsigned long  bg = 0x000000;
   static float  zero[3]  = {0.0, 0.0, 0.0};
   static Colormap  map;
   static unsigned long  fgpix, bgpix;

   dpy = display;
   Width = width;
   Height = height;
   scr = DefaultScreen(dpy);
   screen = ScreenOfDisplay(dpy, scr);
   visual = DefaultVisual(dpy,scr);
   depth = DefaultDepth(dpy,scr);
   map = DefaultColormapOfScreen(screen);

   if (visual->class!=TrueColor)
      die("Display is not TrueColor.\n");

   /* Set background and foreground colors */
   background.red   = (bg >> 16) << 8;
   background.green = (bg & 0xff00);
   background.blue  = (bg & 0xff) << 8;
   background.flags = DoRed | DoGreen | DoBlue;
   foreground.red   = (fg >> 16) << 8;
   foreground.green = (fg & 0xff00);
   foreground.blue  = (fg & 0xff) << 8;
   foreground.flags = DoRed | DoGreen | DoBlue;

   /* Allocate foreground and background colors */
   if (XAllocColor(dpy, map, &background))
      bgpix = background.pixel;
   else {
      fprintf(stderr, "Can't allocate background color\n");
      bgpix = 0;
   }
   if (XAllocColor(dpy, map, &foreground))
      fgpix = foreground.pixel;
   else {
      fprintf(stderr, "Can't allocate foreground color\n");
      fgpix = 0xffffff;
   }

   /* set window attributes */
   xswa.colormap = map;
   xswa.event_mask = ExposureMask | PointerMotionMask | KeyPressMask
                | ButtonPressMask | ButtonReleaseMask | StructureNotifyMask;
   xswa.background_pixel = bgpix;
   xswa.border_pixel = fgpix;
   swa_flags = CWColormap | CWEventMask | CWBackPixel | CWBorderPixel;

   if (width==ScrWidth && height==ScrHeight)
     borderwidth = 0;
   else
     borderwidth = 5;

   /* Create the window */
   wmain = XCreateWindow( dpy, RootWindow(dpy, scr), xpos, ypos,
                        width, height, borderwidth,
		        depth, InputOutput, visual, swa_flags, &xswa);

   if (!wmain)
      die("Unable to open 3-D window");

   /* Set size hints for window manager */
   sizehints.x = xpos;
   sizehints.y = ypos;
   sizehints.width  = width;
   sizehints.height = height;
   sizehints.flags = USSize | USPosition;

   XSetNormalHints(dpy, wmain, &sizehints);
   XSetStandardProperties( dpy, wmain, title, title,
		           None, (char **)NULL, 0, &sizehints);

   /* make gc */
   gcv.foreground = fg;
   gcv.background = bg;
   gc1 = XCreateGC(dpy, wmain, GCForeground | GCBackground, &gcv);

   xgc = XCreateGC(dpy, wmain, 0, NULL );

   /* set font */
   if (fontname[0]==0)
      strcpy( fontname, DEFAULT_FONT );

   if ( !(visfont = XLoadQueryFont( dpy, fontname )) ) {
      printf( "Unable to load font: %s\n", fontname );
      exit(-1);
   }
   XSetFont( dpy, gc1, visfont->fid );
   FontHeight = visfont->ascent + visfont->descent;
   FontDescent = visfont->descent;

   /* set pointer shape */
   XDefineCursor( dpy, wmain, XCreateFontCursor(dpy,XC_top_left_arrow) );
   /* wait for window to be mapped */
   XMapWindow(dpy, wmain);
   do {
      XNextEvent(dpy, &pe);
   } while (pe.type!=Expose);
   XSync(dpy, 0);


   /*Initialize our offscreen pixmap.  This consists of the following
     objects: A buffer == (offscreen pixmap associated with a window),
     Renderable = (buffer made "renderable"), Z-buffer = (initially assoc'd
     with a renderable), gc = (3D graphics context - contains assoc of
     renderable).
   */
   att.update_action = XdbRetained;  /* Don't clear pixmap between redraws */
   att.update_mode = XdbFrequently;  /* Hint to server about # of updates  */
   att.resize_mode = XdbUndefinedBits;
   mask = DBUpdateAction | DBUpdateMode;

   /* Create a Renderable */
   buff0 = XdbCreateBuffer(dpy, wmain, &att, mask, NULL);
   XFDIMakeRenderable(dpy, buff0);
   XFDICreateZbuffer(dpy, buff0);

   /* create XFDI GC */
   gc0 = XFDICreateGc3(dpy, buff0);
   XFDISetVectorFlag(dpy, gc0, XFDIObjectMatrix, XFDIVectorNoHomo);
   XFDISetClipPlane( dpy, gc0, 0, XFDIClipAfterTransform,
                     0.0, 0.0, 1.0, 0.0, 1.0 );
   XFDISetClipLevel( dpy, gc0, (XFDIClipClipPlanes ^ XFDIClipPlane0)
                     | XFDIClipZmin );
   XFDISetRenderable(dpy, gc0, (Drawable) buff0);

   XFDISetSurfaceBackground(dpy, gc0, bgpix);
   XFDISetLineBackground(dpy, gc0, bgpix);

   /* Set surface properties */
   XFDISetSurfaceLightingModel(dpy, gc0, XFDIDiffuse);
   XFDISetSurfaceShading(dpy, gc0, XFDIShadeColor);
   XFDISetTransparencyMode( dpy, gc0, XFDITransparencyFast );
   XFDISetSurfaceTransparency(dpy, gc0, 1.0);

   /* Set the drawing mode - z buffered */
   XFDISetDrawMode(dpy, gc0, XFDIDrawZ);

   /* set depth cueing numbers */
   XFDISetDepthCueing(dpy, gc0, 1.0, -1.0, 0.5, 1.0);
   XFDISetDepthCueingFlag(dpy, gc0, 1 );

   /* Setup light sources */
   XFDISetLight(dpy, gc0, 0, XFDIAmbLight, zero, zero, AmbientColor);
   XFDISetLightStatus(dpy, gc0, 0, XFDIOn);

   XFDISetLightStatus(dpy, gc0, 1, XFDIOn);
   XFDISetLightStatus(dpy, gc0, 2, XFDIOn);

   return wmain;
#endif

#ifdef sgi
   int scr;
   static long Black = 0xff000000;
   char title2[100];

   scr = DefaultScreen(display);

   strcpy( title2, title );  /* uniquify the window title */
   strcat( title2, "   " );

   foreground();

   if (width==ScrWidth && height==ScrHeight) {
      noborder();
      prefposition( 0, ScrWidth-1, 0, ScrHeight-1 );
   }
   else {
      prefposition( xpos, xpos+width, ScrHeight-ypos, ScrHeight-ypos-height );
   }
   winopen( title2 );
   winconstraints();

   /* find X window pointer to the GL window */
   wmain = find_window( display, scr, RootWindow(display,scr), title2 );
   if (!wmain) {
      fprintf(stderr,"Unable to create graphics window\n");
      exit(1);
   }
   wintitle( title );   /* restore window title */

   /* select which X events to receive */
   XSelectInput( display, wmain, ExposureMask | PointerMotionMask
		| KeyPressMask | ButtonPressMask | ButtonReleaseMask
		| StructureNotifyMask | VisibilityChangeMask );

   /* setup 24-bit, double buffer, z-buffered window */
   RGBmode();
   doublebuffer();
   gconfig();
   zbuffer( TRUE );
   zfunction( ZF_LESS );  /* improves transparency quality */
   zfar = getgdesc( GD_ZMAX );

   czclear( Black, zfar );
   swapbuffers();
   czclear( Black, zfar );

   mmode( MVIEWING );     /* multi-matrix mode */

   getsize( &Width, &Height );
   AspectRatio = (float) Width / (float) Height;
   ortho( -1.0, 1.0, -1.0, 1.0, 1.0, -1.0 );  /* this MUST be after mmode()! */
   viewport( 0, Width-1, 0, Height-1);

   lmdef( DEFMATERIAL, 10, 0, mat1 );   /* define material, light, model */
   lmdef( DEFLIGHT, 20, 0, light1 );
   lmdef( DEFLIGHT, 21, 0, light2 );
   lmdef( DEFLMODEL, 30, 0, lmodel );
   lmbind( MATERIAL, 10 );
   lmbind( LIGHT1, 20 );
   lmbind( LIGHT2, 21 );
   lmbind( LMODEL, 30 );

   /* Set font */
   fminit();
   if (fontname[0]==0)
      strcpy( fontname, DEFAULT_FONT );

   visfont = fmfindfont( fontname );
   if (!visfont)
      die("Unable to find font");
   if (fontsize)
      FontHeight = fontsize;
   else
      FontHeight = 20;

   visfont = fmscalefont( visfont, (double) FontHeight );
   fmsetfont( visfont );
   FontDescent = getdescender();

   /* init front clipping plane */
   front = 1;

   /* box clipping plane */
   clipplane( 1, CP_DEFINE, xminplane );
   clipplane( 1, CP_OFF, xminplane );
   clipplane( 2, CP_DEFINE, xmaxplane );
   clipplane( 2, CP_OFF, xmaxplane );
   clipplane( 3, CP_DEFINE, yminplane );
   clipplane( 3, CP_OFF, yminplane );
   clipplane( 4, CP_DEFINE, ymaxplane );
   clipplane( 4, CP_OFF, ymaxplane );

   /* screen door transparency patterns */
   defpattern( 1, 16, screen1 );
   defpattern( 2, 16, screen2 );
   defpattern( 3, 16, screen3 );
   setpattern(0);

   subpixel(TRUE);   /* for faster rendering on VGX */
   return wmain;
#endif
#ifdef ibm
   int scr;
   static long Black = 0xff000000;
   char title2[100];

   scr = DefaultScreen(display);

   strcpy( title2, title );  /* uniquify the window title */
   strcat( title2, "   " );

   prefposition( xpos, xpos+width, ScrHeight-ypos, ScrHeight-ypos-height );
   winopen( title2 );
   winconstraints();

   /* find X window pointer to the GL window */
   wmain = find_window( display, scr, RootWindow(display,scr), title2 );
   if (!wmain) {
      fprintf(stderr,"Unable to create graphics window\n");
      exit(1);
   }
   wintitle( title );   /* restore window title */

   /* select which X events to receive */
   XSelectInput( display, wmain, ExposureMask | PointerMotionMask
		| KeyPressMask | ButtonPressMask | ButtonReleaseMask
		| StructureNotifyMask | VisibilityChangeMask );

   /* setup 24-bit, double buffer, z-buffered window */
   RGBmode();
   doublebuffer();
   gconfig();
   zbuffer( TRUE );
   zfar = getgdesc( GD_ZMAX );

   czclear( Black, zfar );
   swapbuffers();
   czclear( Black, zfar );

   mmode( MVIEWING );     /* multi-matrix mode */

   getsize( &Width, &Height );
   AspectRatio = (float) Width / (float) Height;
   ortho( -1.0, 1.0, -1.0, 1.0, 1.0, -1.0 );  /* this MUST be after mmode()! */
   viewport( 0, Width-1, 0, Height-1);

   lmdef( DEFMATERIAL, 10, 0, mat1 );   /* define material, light, model */
   lmdef( DEFLIGHT, 20, 0, light1 );
   lmdef( DEFLIGHT, 21, 0, light2 );
   lmdef( DEFLMODEL, 30, 0, lmodel );
   lmbind( MATERIAL, 10 );
   lmbind( LIGHT1, 20 );
   lmbind( LIGHT2, 21 );
   lmbind( LMODEL, 30 );

   /* Set font */
   FontHeight = getheight();
   FontDescent = getdescender();

   /* init front clipping plane */
   front = 1;

   /* screen door transparency patterns */
   defpattern( 1, 16, screen1 );
   defpattern( 2, 16, screen2 );
   defpattern( 3, 16, screen3 );
   setpattern(0);

   return wmain;
#endif
#ifdef foo
   /* open a 3-D rendering window and return the X window identifier */
#endif
}




/*** clear_window *****************************************************
   Clear the graphics window.  This is called prior to rendering
   a frame.
**********************************************************************/
void clear_window( bgcolor )
unsigned long bgcolor;
{
#ifdef stellar
   Window  gr;              /* Returned by */
   int  gx, gy, gbw, gd;    /* XGetGeometry */

   /* Get current info about window (size, etc.) */
   XGetGeometry(dpy, wmain, &gr, &gx, &gy, &Width, &Height, &gbw, &gd);

   XFDIClear(dpy, gc0, bgcolor,
                 (unsigned long) 0x0, (XRectangle *) NULL);
   XFDISetLineBackground( dpy, gc0, bgcolor );

   AspectRatio = 1.0;
#endif
#ifdef sgi_or_ibm
   getsize( &Width, &Height );
   AspectRatio = (float) Width / (float) Height;
   viewport( 0, Width-1, 0, Height-1);

   czclear( bgcolor, zfar );
#endif
#ifdef foo
   /* clear the 3-D window to bgcolor before rendering into it */
#endif
}




/*** update_window ****************************************************
   Update the 3-D window.  This called when finished rendering a frame.
**********************************************************************/
void update_window()
{
#ifdef stellar
   if (pretty_flag)
     XdbUpdateWindows(dpy, &aabuff0, NULL, 1);
   else
     XdbUpdateWindows(dpy, &buff0, NULL, 1);
#endif
#ifdef sgi
   if (pretty_flag) {
      acbuf( AC_RETURN, 1.0/(float) AA_PASSES );
   }
   swapbuffers();
#endif
#ifdef ibm
   swapbuffers();
#endif
#ifdef foo
   /* copy back buffer to front buffer */
#endif
}



/*** set_2d ***********************************************************
   This is to be called prior to any 2-D rendering calls.
**********************************************************************/
void set_2d()
{
#ifdef sgi_or_ibm
   loadmatrix( IdMat );
   getsize( &Width, &Height );
   ortho2( -0.5, Width-0.5, -0.5, Height-0.5 );
#endif
#ifdef foo
   /* do anything needed to switch from 3-D to 2-D rendering */
#endif
}



/*** set_3d ***********************************************************
   This is to be called prior to any 3-D rendering calls.
   Input:  ctm - the current transformation matrix.
**********************************************************************/
void set_3d( ctm, perspec )
Matrix ctm;
int perspec;
{
#ifdef stellar
   static float zero[3] = {0.0, 0.0, 0.0};
   float dir[3];
   float mag;

   mat_copy( Ctm, ctm );
   mat_inv( CtmInv, ctm );

   XFDISetMatrix(dpy, gc0, XFDIObjectMatrix, ctm, XFDIMatrix4x4);

   dir[0] = ctm[0][2];   dir[1] = ctm[1][2];   dir[2] = ctm[2][2];
   mag = (float) sqrt( dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2] );
   dir[0] /= mag;
   dir[1] /= mag;
   dir[2] /= mag;
   XFDISetLight(dpy, gc0, 1, XFDIDirLight, zero, dir, LightColor );

   dir[0] = -dir[0];   dir[1] = -dir[1];   dir[2] = -dir[2];
   XFDISetLight(dpy, gc0, 2, XFDIDirLight, zero, dir, LightColor );
#endif
#ifdef sgi_or_ibm
   perspec = 0;

   mat_copy( Ctm, ctm );
   mat_inv( CtmInv, ctm );

   if (perspec) {
      perspective( 400, 1.0, -1.0, -100.0 );
/*      loadmatrix( IdMat );
      lookat( 0.0,0.0,8.0,  0.0,0.0,0.0, 0.0 );*/
      Ctm[3][2] = 5.0;
      loadmatrix( Ctm );

/*      lookat( 0.0, 0.0, 10.0,  0.0, 0.0, 0.0,  0.0 );*/
   }
   else {
      /* Note reversed Z! */
      ortho( -AspectRatio, AspectRatio, -1.0, 1.0, front, -1.0 );
      loadmatrix( ctm );
   }
#endif
#ifdef foo
   /* get ready for 3-D rendering */
#endif
}



/*** set_color ********************************************************
   Set the current drawing color.  The color is specified in packed
   ARGB format; 1-byte each of Alpha, Red, Green, Blue where Alpha
   is the most-significant byte and blue is the least-signficant byte.
**********************************************************************/
void set_color( c )
unsigned long c;
{
#ifdef stellar
   XFDISetSurfaceColor( dpy, gc0, c );
   XFDISetLineColor( dpy, gc0, c );
   CurTrans = (float) ((c>>24) & 0xff) / 255.0;
   XFDISetSurfaceTransparency( dpy, gc0, CurTrans );
   XSetForeground( dpy, gc1, c );
#endif
#ifdef sgi_or_ibm
   CurColor = FLIP( c );
   CurTrans = c & 0xff000000;
   if (screendoor) {
      CurTrans = 3 - (CurTrans>>24) / 64;
   }
   cpack( CurColor );
#endif
#ifdef foo
   /* set current drawing color */
#endif
}



/*** set_depthcue *****************************************************
   Set the line depth cueing flag on or off.
**********************************************************************/
void set_depthcue( onoff )
int onoff;
{
#ifdef stellar
   if (onoff)
      XFDISetDepthCueingFlag(dpy, gc0, 1 );
   else
      XFDISetDepthCueingFlag(dpy, gc0, 0 );
#endif
#ifdef sgi
   short r, g, b;
   if (onoff) {
      r = (CurColor      ) & 0xff;
      g = (CurColor >> 8 ) & 0xff;
      b = (CurColor >> 16) & 0xff;
      depthcue( TRUE );
      lRGBrange( r/5, g/5, b/5,  r, g, b, 0, zfar );
   }
   else {
      depthcue( FALSE );
   }
#endif
#ifdef ibm
   short r, g, b;
   if (onoff) {
      r = (CurColor      ) & 0xff;
      g = (CurColor >> 8 ) & 0xff;
      b = (CurColor >> 16) & 0xff;
      depthcue( TRUE );
      lRGBrange( r/3, g/3, b/3,  r, g, b, 0, zfar );
   }
   else {
      depthcue( FALSE );
   }
#endif
#ifdef foo
   /* if your system supports depth cued lines, turn it on or off here */
#endif
}



/*** set_line_width ***************************************************
   Set the width of lines.
   Input:  w - width in pixels.
**********************************************************************/
void set_line_width( w )
float w;
{
#ifdef stellar
   linewidth = w;
   XFDISetLineWidth( dpy, gc0, w );
#endif
#ifdef sgi
   linewidthf( w );
#endif
#ifdef ibm
/*   linewidthf( w );*/
#endif
#ifdef foo
   /* if your system supports wide line drawing, set it here. */
#endif
}



/*** set_frontclip ****************************************************
   Set the position of the user-defined front clipping plane.
   Input:  d - front clipping plane z position.
**********************************************************************/
void set_frontclip( d )
float d;
{
#ifdef stellar
   XFDISetClipPlane(dpy, gc0, 0, XFDIClipAfterTransform,
		    0.0, 0.0, 1.0, 0.0, d );
#endif
#ifdef sgi_or_ibm
   if (d>1.0)
      front = 1.0;
   else if (d<-0.95)
      front = -0.95;
   else
      front = d;
   /* set_3d() will put new front clipping plane into effect */
#endif
#ifdef foo
   /* set front clipping plane distance here */
#endif
}




/*** set_box_clipping *************************************************
   Enable or disable clipping of graphics to a box.
   Input:  onoff - 1 = enable, 0 = disable.
           xmin, xmax, ymin, ymax - clipping bounds in graphics space,
                                    note no Z clipping is done.
**********************************************************************/
void set_box_clipping( onoff, xmin, xmax, ymin, ymax )
int onoff;
float xmin, xmax, ymin, ymax;
{
#ifdef stellar
   unsigned long planes;

   if (onoff) {

      XFDISetClipPlane( dpy, gc0, 1, XFDIClipAfterTransform,
	             CtmInv[0][0], CtmInv[1][0], CtmInv[2][0], -xmin, 0.0 );
      XFDISetClipPlane( dpy, gc0, 2, XFDIClipAfterTransform,
	            -CtmInv[0][0], -CtmInv[1][0], -CtmInv[2][0], xmax, 0.0 );
      XFDISetClipPlane( dpy, gc0, 3, XFDIClipAfterTransform,
	             CtmInv[0][1], CtmInv[1][1], CtmInv[2][1], -ymin, 0.0 );
      XFDISetClipPlane( dpy, gc0, 4, XFDIClipAfterTransform,
	            -CtmInv[0][1], -CtmInv[1][1], -CtmInv[2][1], ymax, 0.0 );

      planes = XFDIClipZmin | (XFDIClipClipPlanes ^
	    (XFDIClipPlane0|XFDIClipPlane1|XFDIClipPlane2|
	     XFDIClipPlane3|XFDIClipPlane4));
      XFDISetClipLevel( dpy, gc0, planes );
   }
   else {
      planes = (XFDIClipClipPlanes ^ XFDIClipPlane0) | XFDIClipZmin;
      XFDISetClipLevel( dpy, gc0, planes );
   }
#endif
#ifdef sgi
   if (onoff) {
      xminplane[0]=  1.0;  xminplane[1]=xminplane[2]=0.0;  xminplane[3]= -xmin;
      xmaxplane[0]= -1.0;  xmaxplane[1]=xmaxplane[2]=0.0;  xmaxplane[3]=  xmax;
      yminplane[1]=  1.0;  yminplane[0]=yminplane[2]=0.0;  yminplane[3]= -ymin;
      ymaxplane[1]= -1.0;  ymaxplane[0]=ymaxplane[2]=0.0;  ymaxplane[3]=  ymax;

      clipplane( 1, CP_DEFINE, xminplane );  clipplane( 1, CP_ON, xminplane );
      clipplane( 2, CP_DEFINE, xmaxplane );  clipplane( 2, CP_ON, xmaxplane );
      clipplane( 3, CP_DEFINE, yminplane );  clipplane( 3, CP_ON, yminplane );
      clipplane( 4, CP_DEFINE, ymaxplane );  clipplane( 4, CP_ON, ymaxplane );
   }
   else {
      clipplane( 1, CP_OFF, xminplane );
      clipplane( 2, CP_OFF, xmaxplane );
      clipplane( 3, CP_OFF, yminplane );
      clipplane( 4, CP_OFF, ymaxplane );
   }
#endif
#ifdef ibm
   /* nothing */
#endif
#ifdef foo
   /* nothing */
#endif
}




/*** set_pretty *******************************************************
   Set pretty rendering mode on or off.
   Stellar:  This function is used to select antialiasing and multi-pass
             transparency.
   SGI:  This function selects line antialiasing.
   Input: onoff - 1 = on, 0 = off.
**********************************************************************/
void set_pretty( onoff )
{
#ifdef stellar
   if (onoff && !pretty_flag) {
      /* turn on */
      /* Create antialiasing buffer */
      aabuff0 = XdbCreateBuffer(dpy, wmain, &att, mask );
      XFDIMakeRenderable(dpy, aabuff0);
      XFDISetBuffer(dpy, gc0, aabuff0, XFDIPmapFilterI );
      /* XFDISetFilter(dpy, gc0, XFDIFilterGeneral ); */
      XFDISetFilter(dpy, gc0, XFDIFilterFilter );

      /* Create multi-pass transparency buffers */
      sorta = XdbCreateBuffer( dpy, wmain, &att, mask );
      sorti = XdbCreateBuffer( dpy, wmain, &att, mask );
      sortz = XdbCreateBuffer( dpy, wmain, &att, mask );
      XFDIMakeRenderable(dpy, sorta);
      XFDIMakeRenderable(dpy, sorti);
      XFDIMakeRenderable(dpy, sortz);
      XFDISetBuffer( dpy, gc0, sorta, XFDIPmapSortA );
      XFDISetBuffer( dpy, gc0, sorti, XFDIPmapSortI );
      XFDISetBuffer( dpy, gc0, sortz, XFDIPmapSortZ );

      XFDISetTransparencyMode( dpy, gc0, XFDITransparencyHQ );

      /* make busy cursor */
      XDefineCursor( dpy, wmain, XCreateFontCursor(dpy,XC_watch) );
      pretty_flag = 1;
   }
   else if (!onoff && pretty_flag) {
      /* turn off */
      XdbDestroyBuffer(dpy, aabuff0);
      XdbDestroyBuffer(dpy, sorta);
      XdbDestroyBuffer(dpy, sorti);
      XdbDestroyBuffer(dpy, sortz);
      XFDISetFilter(dpy, gc0, XFDIFilterNone );
      XFDISetTransparencyMode( dpy, gc0, XFDITransparencyFast );
      /* regular cursor */
      XDefineCursor( dpy, wmain, XCreateFontCursor(dpy,XC_top_left_arrow) );
      pretty_flag = 0;
   }
#endif
#ifdef sgi
   if (onoff) {
      /* turn on */
      /*subpixel(TRUE);*/
      glcompat( GLC_OLDPOLYGON, 0 );
      acsize(16);
      gconfig();
      /* make busy cursor */
      curstype( C16X1 );
      defcursor( 1, hourglass );
      setcursor( 1 );
      pretty_flag = 1;
   }
   else {
      /* turn off */
      /*subpixel(FALSE);*/
      glcompat( GLC_OLDPOLYGON, 1 );
      acsize(0);
      gconfig();
      /* regulay cursor */
      setcursor( 0 );
      pretty_flag = 0;
   }
#endif
#ifdef ibm
   /* always off */
   pretty_flag = 0;
#endif
#ifdef foo
   pretty_flag = 0;
#endif
}



/*** start_aa_pass ****************************************************
   Start antialiasing pass n, where n is in [0..8].
**********************************************************************/
void start_aa_pass( n )
int n;
{
#ifdef stellar
   if (pretty_flag) {
      XFDISetFilterOffset(dpy, gc0, xoffsets[n],yoffsets[n] );
   }
#endif
#ifdef sgi
   float w, h;

   if (pretty_flag) {
      if (n==0) {
	 /* clear ACC buffer */
	 acbuf( AC_CLEAR, 0.0 );
      }
      w = xoffsets[n]/Width;
      h = yoffsets[n]/Height;
      ortho( -AspectRatio+w, AspectRatio+w, -1.0+h, 1.0+h, front, -1.0 );
      loadmatrix( Ctm );
   }
#endif
#ifdef ibm
/*   ortho( -AspectRatio, AspectRatio, -1.0, 1.0, front, -1.0 );
   loadmatrix( Ctm ); */
#endif
#ifdef foo
   /* nothing */
#endif
}




/*** end_aa_pass ******************************************************
   End an antialiasing pass n, where n is in [0..8].
**********************************************************************/
void end_aa_pass( n )
int n;
{
#ifdef stellar
   if (pretty_flag) {
      XFDIFilterBlend( dpy, gc0, blendcoeffs[n],0,0, Width, Height );
   }
#endif
#ifdef sgi
   if (pretty_flag) {
      acbuf( AC_ACCUMULATE, 1.0 );
   }
#endif
#ifdef foo
   /* nothing */
#endif
}



/*** start_hqtrans_pass ***********************************************
   Start a high-quality transparency rendering pass.
**********************************************************************/
void start_hqtrans_pass()
{
#ifdef stellar
   if (pretty_flag) {
      XFDISetTransparencyMode( dpy, gc0, XFDITransparencyHQ );

      XSetForeground(dpy,xgc,0xffffffff);
      XFillRectangle(dpy,sortz,xgc,0,0,Width,Height);
      XSetForeground(dpy,xgc,0x0);
      XFillRectangle(dpy,sorti,xgc,0,0,Width,Height);
      XFillRectangle(dpy,sorta,xgc,0,0,Width,Height);
   }
#endif
#ifdef foo
   /* nothing */
#endif
}



/*** end_hqtrans_pass *************************************************
   End a high-quality transparency rendering pass.
   Return:  number of unresolved pixels (used to determine when enough
            rendering passes have been made.
**********************************************************************/
int end_hqtrans_pass()
{
#ifdef stellar
   int pixels = 0;
   if (pretty_flag) {
      XFDITransBlend( dpy, gc0, 0,0, Width, Height, &pixels );
      XFDISetTransparencyMode( dpy, gc0, XFDITransparencyFast );
   }
   return pixels;
#endif
#ifdef sgi_or_ibm
   return 0;
#endif
#ifdef foo
   /* nothing */
   return 0;
#endif
}




/*** polytrinorm ******************************************************
   Draw a polytriangle strip with normals.
**********************************************************************/
void polytrinorm( vert, norm, n )
float vert[][3];
float norm[][3];
int n;
{
#ifdef stellar
   if (CurTrans>0.0) {
      XFDIVnormPolyTri( dpy, gc0, vert, norm, n );
   }
#endif
#ifdef sgi
   int i;

   lmcolor( LMC_AD );
   cpack( CurColor );
   /* enable transparency */
   if (screendoor)
     setpattern( CurTrans );
   else
     blendfunction( BF_SA, BF_MSA );

#ifdef WIREFRAME
   /* draw wireframe triangles instead of solid shaded */
   for (i=2;i<n;i++) {
      bgnline();
      n3f( norm[i-1] );  v3f( vert[i-1] );
      n3f( norm[i] );    v3f( vert[i] );
      n3f( norm[i-2] );  v3f( vert[i-2] );
      endline();
   }
#else
   bgntmesh();
   for (i=0;i<n;i++) {
      n3f( norm[i] );
      v3f( vert[i] );
   }
   endtmesh();
#endif

   /* turn off transparency */
   if (screendoor)
     setpattern(0);
   else
     blendfunction( BF_ONE, BF_ZERO );
#endif
#ifdef ibm
   int i;

   lmcolor( LMC_AD );
   cpack( CurColor );
   /* enable transparency */
   if (screendoor)
     setpattern( CurTrans );

   /* Due to strangeness in the IBM hardware and their version of GL, */
   /* we must break the polytriangle strip into individual triangles  */
   /* and test to make sure the normal vectors are correct.  Also, we */
   /* must cull out triangles with zero area.                         */
   for (i=2; i<n; i++) {
      /* calculate triangle surface normal */
      float v1[3], v2[3], cross[3];
      v1[0] = vert[i][0] - vert[i-2][0];
      v1[1] = vert[i][1] - vert[i-2][1];
      v1[2] = vert[i][2] - vert[i-2][2];
      v2[0] = vert[i-1][0] - vert[i][0];
      v2[1] = vert[i-1][1] - vert[i][1];
      v2[2] = vert[i-1][2] - vert[i][2];
      cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
      cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
      cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
      /* check for zero area */
      if (cross[0]==0.0 && cross[1]==0.0 && cross[2]==0.0)
	continue;

      bgntmesh();

      /* vertex & normal 1 */
      if (DOT(cross,norm[i-2])<0.0) {
	 /* flip normal */
	 float nn[3];
	 nn[0] = -norm[i-2][0];
	 nn[1] = -norm[i-2][1];
	 nn[2] = -norm[i-2][2];
	 n3f( nn );
      }
      else
	n3f( norm[i-2] );
      v3f(vert[i-2]);

      /* vertex & normal 2 */
      if (DOT(cross,norm[i-1])<0.0) {
	 /* flip normal */
	 float nn[3];
	 nn[0] = -norm[i-1][0];
	 nn[1] = -norm[i-1][1];
	 nn[2] = -norm[i-1][2];
	 n3f( nn );
      }
      else
	n3f( norm[i-1] );
      v3f(vert[i-1]);

      /* vertex & normal 3 */
      if (DOT(cross,norm[i])<0.0) {
	 /* flip normal */
	 float nn[3];
	 nn[0] = -norm[i][0];
	 nn[1] = -norm[i][1];
	 nn[2] = -norm[i][2];
	 n3f( nn );
      }
      else
	n3f( norm[i] );
      v3f(vert[i]);

      endtmesh();
   }

   /* Unbind and rebind the material to get around another */
   /* strange IBM GL bug */
   lmbind( MATERIAL, 0 );
   lmbind( MATERIAL, 10 );

   /* turn off transparency */
   setpattern(0);
#endif
#ifdef foo
   /* draw polytriangle strip with normal vectors at vertices */
#endif
}



/*** polytri **********************************************************
   Draw a polytriangle strip without shading.
**********************************************************************/
void polytri( vert, n )
float vert[][3];
int n;
{

#ifdef stellar
   if (CurTrans>0.0) {
      XFDIPolyTri( dpy, gc0, vert, n );
   }
#endif

#ifdef sgi
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );
   /* enable transparency */
   if (screendoor)
     setpattern( CurTrans );
   else
     blendfunction( BF_SA, BF_MSA );

   bgntmesh();
   for (i=0;i<n;i++) {
      v3f( vert[i] );
   }
   endtmesh();

   /* turn off transparency */
   if (screendoor)
     setpattern(0);
   else
     blendfunction( BF_ONE, BF_ZERO );

#endif
#ifdef ibm
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );
   setpattern( CurTrans );

   bgntmesh();
   for (i=0;i<n;i++) {
      v3f( vert[i] );
   }
   endtmesh();
   setpattern(0);
#endif   
#ifdef foo
   /* draw flat-shaded polytriangle strip */
#endif
}



/*** polyline *********************************************************
   Draw a 3-D poly line.
**********************************************************************/
void polyline( vert, n )
float vert[][3];
int n;
{

#ifdef stellar
   XFDISetLineWidth( dpy, gc0, linewidth );
   XFDIPolyLine( dpy, gc0, vert, n );
#endif
#ifdef sgi
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );
/*
   if (pretty_flag) {
      subpixel(TRUE);
      linesmooth( SML_ON | SML_SMOOTHER | SML_END_CORRECT );
      blendfunction( BF_SA, BF_MSA );
   }
*/

   bgnline();
   for (i=0;i<n;i++) {
      v3f( vert[i] );
   }
   endline();

/*
   subpixel(FALSE);
   linesmooth( SML_OFF );
   blendfunction( BF_ONE, BF_ZERO );
*/
#endif
#ifdef ibm
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );

   bgnline();
   for (i=0;i<n;i++) {
      v3f( vert[i] );
   }
   endline();
#endif
#ifdef foo
   /* draw a 3-D polyline */
#endif
}




/*** disjointpolyline *************************************************
   Draw a series of disjoint 3-D lines.
**********************************************************************/
void disjointpolyline( vert, n )
float vert[][3];
int n;
{

#ifdef stellar
   XFDISetLineWidth( dpy, gc0, linewidth );
   XFDIDisjointPolyLine( dpy, gc0, vert, n );
#endif
#ifdef sgi
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );
/*
   if (pretty_flag) {
      subpixel(TRUE);
      linesmooth( SML_ON | SML_SMOOTHER | SML_END_CORRECT );
      blendfunction( BF_SA, BF_MSA );
   }
*/

   for (i=0;i<n;i+=2) {
      bgnline();
      v3f( vert[i] );
      v3f( vert[i+1] );
      endline();
   }

/*
   subpixel(FALSE);
   linesmooth( SML_OFF );
   blendfunction( BF_ONE, BF_ZERO );
*/
#endif
#ifdef ibm
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );

   for (i=0;i<n;i+=2) {
      bgnline();
      v3f( vert[i] );
      v3f( vert[i+1] );
      endline();
   }
#endif
#ifdef foo
   /* draw 3-D disjoint polyline */
#endif
}





/*** quadmesh *********************************************************
   Draw a quadrilateral mesh with colors at each vertex.
**********************************************************************/
void quadmesh( vert, color, rows, cols )
float vert[][3];
unsigned int color[];
int rows, cols;
{
#ifdef stellar
   if (CurTrans>0.0) {
      /* actually render as a triangle mesh */
      XFDIMeshTriVcolors( dpy, gc0, vert, color, rows, cols );
   }
#endif
#ifdef sgi
   int i, j, base1, base2, c;

   lmcolor( LMC_COLOR );
   /* enable transparency blending */
   if (screendoor)
     setpattern( CurTrans );
   else
     blendfunction( BF_SA, BF_MSA );

   /* break mesh into strips */
   for (i=0;i<rows-1;i++) {
      base1 = i * cols;
      base2 = (i+1) * cols;
      bgnqstrip();
      for (j=0;j<cols;j++) {
	 cpack( FLIP( color[base1+j] ) );
	 v3f( vert[base1+j] );
	 cpack( FLIP( color[base2+j] ) );
	 v3f( vert[base2+j] );
      }
      endqstrip();
   }

   /* turn off transparency */
   if (screendoor)
     setpattern(0);
   else
     blendfunction( BF_ONE, BF_ZERO );
#endif
#ifdef ibm
   int i, j, base1, base2, c;

   lmcolor( LMC_COLOR );

   setpattern( CurTrans );

   /* break mesh into strips */
   for (i=0;i<rows-1;i++) {
      base1 = i * cols;
      base2 = (i+1) * cols;
      bgntmesh();
      for (j=0;j<cols;j++) {
	 cpack( FLIP( color[base1+j] ) );
	 v3f( vert[base1+j] );
	 cpack( FLIP( color[base2+j] ) );
	 v3f( vert[base2+j] );
      }
      endtmesh();
   }

   setpattern(0);
#endif
#ifdef foo
   /* draw quadrilateral mesh with vertex colors */
#endif
}



/*** quadmeshnorm *****************************************************
   Draw a quadrilateral mesh with normals and colors at each vertex.
**********************************************************************/
void quadmeshnorm( vert, norm, color, rows, cols )
float vert[][3];
float norm[][3];
unsigned int color[];
int rows, cols;
{
#ifdef stellar
   if (CurTrans>0.0) {
      /* actually render as a triangle mesh */
      XFDIVnormMeshTriVcolors( dpy, gc0, vert, norm, color, rows, cols );
   }
#endif
#ifdef sgi
   int i, j, base1, base2, c;

   lmcolor( LMC_AD );
   /* enable transparency blending */
   if (screendoor)
     setpattern( CurTrans );
   else
     blendfunction( BF_SA, BF_MSA );

   /* break mesh into strips */
   for (i=0;i<rows-1;i++) {
      base1 = i * cols;
      base2 = (i+1) * cols;
      bgnqstrip();
      for (j=0;j<cols;j++) {
	 cpack( FLIP( color[base1+j] ) );
	 n3f( norm[base1+j] );
	 v3f( vert[base1+j] );
	 cpack( FLIP( color[base2+j] ) );
	 n3f( norm[base2+j] );
	 v3f( vert[base2+j] );
      }
      endqstrip();
   }

   /* turn off transparency */
   if (screendoor)
     setpattern(0);
   else
     blendfunction( BF_ONE, BF_ZERO );
#endif
#ifdef ibm
   int i, j, base1, base2, c;

   lmcolor( LMC_AD );

   setpattern( CurTrans );

   /* break mesh into strips */
   for (i=0;i<rows-1;i++) {
      base1 = i * cols;
      base2 = (i+1) * cols;
      bgntmesh();
      for (j=0;j<cols;j++) {
	 cpack( FLIP( color[base1+j] ) );
	 n3f( norm[base1+j] );
	 v3f( vert[base1+j] );
	 cpack( FLIP( color[base2+j] ) );
	 n3f( norm[base2+j] );
	 v3f( vert[base2+j] );
      }
      endtmesh();
   }

   /* Unbind and rebind the material to get around another */
   /* strange IBM GL bug */
   lmbind( MATERIAL, 0 );
   lmbind( MATERIAL, 10 );

   setpattern(0);
#endif
#ifdef foo
   /* draw quadrilateral mesh with vertex colors and normals */
#endif
}



/*** polyline2d *******************************************************
   Draw a 2-D poly line.  Coordinates are in pixels with the origin
   in the upper-left corner of the window.  NOTE:  vertices are shorts.
**********************************************************************/
void polyline2d( vert, n )
short vert[][2];
int n;
{
#ifdef stellar
   XDrawLines( dpy, buff0, gc1, vert, n, CoordModeOrigin );
#endif
#ifdef sgi_or_ibm
   short v[2];
   int i;

   lmcolor( LMC_COLOR );
   cpack( CurColor );
   bgnline();
   for (i=0;i<n;i++) {
      v[0] = vert[i][0];
      v[1] = Height - vert[i][1];
      v2s( v );
   }
   endline();
#endif
#ifdef foo
   /* draw 2-D polyline */
#endif
}


/*** draw_rect ********************************************************
   Draw a rectangle in 2-D.
**********************************************************************/
void draw_rect( x1, y1, x2, y2 )
int x1, y1, x2, y2;
{
   short v[5][2];

   v[0][0] = x1;  v[0][1] = y1;
   v[1][0] = x2;  v[1][1] = y1;
   v[2][0] = x2;  v[2][1] = y2;
   v[3][0] = x1;  v[3][1] = y2;
   v[4][0] = x1;  v[4][1] = y1;
   polyline2d( v, 5 );
}



/*** draw_text ********************************************************
   Draw a text string.  Position coordinates are in pixels with the
   origin being in the upper-left corner.
**********************************************************************/
void draw_text( xpos, ypos, str )
int xpos, ypos;
char *str;
{
#ifdef stellar
   XDrawString( dpy, buff0, gc1, xpos, ypos, str, strlen(str) );
#endif
#ifdef sgi
   /*cmov2i( xpos, Height-ypos );*/
   /*charstr( str );*/
   cmovi( xpos, Height-ypos, 1.0 );
   fmprstr( str );
#endif
#ifdef ibm
   cmov2i( xpos, Height-ypos );
   charstr( str );
#endif
#ifdef foo
   /* draw a text string */
#endif
}



/*** text_width *******************************************************
   Return the width of a text string in pixels.
**********************************************************************/
int text_width( str )
char *str;
{
   int dir, ascent, descent;
   XCharStruct overall;

#ifdef stellar
   XTextExtents(visfont, str, strlen(str), &dir, &ascent, &descent, &overall);
   return overall.width;
#endif
#ifdef sgi
   return fmgetstrwidth( visfont, str );
#endif
#ifdef ibm
   return strwidth( str );
#endif
#ifdef foo
   /* return the width in pixels of a text string */
#endif
}
